import { __rest } from 'tslib';
import 'reflect-metadata';
import { BinarySet, NumberValueSet } from '@aws/dynamodb-auto-marshaller';
import { DynamoDbSchema, DynamoDbTable } from '@aws/dynamodb-data-mapper';
import { v4 } from 'uuid';

const METADATA_TYPE_KEY = 'design:type';

/**
 * Declare a property in a TypeScript class to be part of a DynamoDB schema.
 * Meant to be used as a property decorator in conjunction with TypeScript's
 * emitted type metadata. If used with in a project compiled with the
 * `emitDecoratorMetadata` option enabled, the type will infer most types from
 * the TypeScript source.
 *
 * Please note that TypeScript does not emit any metadata about the type
 * parameters supplied to generic types, so `Array<string>`, `[number, string]`,
 * and `MyClass[]` are all exposed as `Array` via the emitted metadata. Without
 * additional metadata, this annotation will treat all encountered arrays as
 * collections of untyped data. You may supply either a `members` declaration or
 * a `memberType` declaration to direct this annotation to treat a property as a
 * tuple or typed list, respectively.
 *
 * Member type declarations are required for maps and sets.
 *
 * @see https://www.typescriptlang.org/docs/handbook/decorators.html
 * @see https://www.typescriptlang.org/docs/handbook/compiler-options.html
 * @see https://github.com/Microsoft/TypeScript/issues/2577
 *
 * @example
 *  export class MyClass {
 *      @attribute()
 *      id: string;
 *
 *      @attribute()
 *      subdocument?: MyOtherClass;
 *
 *      @attribute()
 *      untypedCollection?: Array<any>;
 *
 *      @attribute({memberType: {type: 'String'}})
 *      listOfStrings?: Array<string>;
 *
 *      @attribute({members: [{type: 'Boolean', type: 'String'}]})
 *      tuple?: [boolean, string];
 *
 *      @attribute({memberType: {type: 'String'}})
 *      mapStringString?: Map<string, string>;
 *
 *      @attribute()
 *      binary?: Uint8Array;
 *  }
 */
function attribute(parameters = {}) {
    return (target, propertyKey) => {
        if (!Object.prototype.hasOwnProperty.call(target, DynamoDbSchema)) {
            Object.defineProperty(target, DynamoDbSchema, // TypeScript complains about the use of symbols here, though it should be allowed
            { value: deriveBaseSchema(target) });
        }
        const schemaType = metadataToSchemaType(Reflect.getMetadata(METADATA_TYPE_KEY, target, propertyKey), parameters);
        if ((schemaType.keyType ||
            schemaType.indexKeyConfigurations) &&
            [
                'Binary',
                'Custom',
                'Date',
                'Number',
                'String',
            ].indexOf(schemaType.type) < 0) {
            throw new Error(`Properties of type ${schemaType.type} may not be used as index or table keys. If you are relying on automatic type detection and have encountered this error, please ensure that the 'emitDecoratorMetadata' TypeScript compiler option is enabled. Please see https://www.typescriptlang.org/docs/handbook/decorators.html#metadata for more information on this compiler option.`);
        }
        target[DynamoDbSchema][propertyKey] = schemaType;
    };
}
function deriveBaseSchema(target) {
    if (target && typeof target === 'object') {
        const prototype = Object.getPrototypeOf(target);
        if (prototype) {
            return Object.assign({}, deriveBaseSchema(prototype), Object.prototype.hasOwnProperty.call(prototype, DynamoDbSchema)
                ? prototype[DynamoDbSchema]
                : {});
        }
    }
    return {};
}
function metadataToSchemaType(ctor, declaration) {
    let { type } = declaration, rest = __rest(declaration, ["type"]);
    if (type === undefined) {
        if (ctor) {
            if (ctor === String) {
                type = 'String';
            }
            else if (ctor === Number) {
                type = 'Number';
            }
            else if (ctor === Boolean) {
                type = 'Boolean';
            }
            else if (ctor === Date || ctor.prototype instanceof Date) {
                type = 'Date';
            }
            else if (ctor === BinarySet ||
                ctor.prototype instanceof BinarySet) {
                type = 'Set';
                rest.memberType = 'Binary';
            }
            else if (ctor === NumberValueSet ||
                ctor.prototype instanceof NumberValueSet) {
                type = 'Set';
                rest.memberType = 'Number';
            }
            else if (ctor === Set || ctor.prototype instanceof Set) {
                type = 'Set';
                if (!('memberType' in rest)) {
                    throw new Error('Invalid set declaration. You must specify a memberType');
                }
            }
            else if (ctor === Map || ctor.prototype instanceof Map) {
                type = 'Map';
                if (!('memberType' in rest)) {
                    throw new Error('Invalid map declaration. You must specify a memberType');
                }
            }
            else if (ctor.prototype[DynamoDbSchema]) {
                type = 'Document';
                rest.members = ctor.prototype[DynamoDbSchema];
                rest.valueConstructor = ctor;
            }
            else if (isBinaryType(ctor)) {
                type = 'Binary';
            }
            else if (ctor === Array || ctor.prototype instanceof Array) {
                if ('members' in declaration) {
                    type = 'Tuple';
                }
                else if ('memberType' in declaration) {
                    type = 'List';
                }
                else {
                    type = 'Collection';
                }
            }
            else {
                type = 'Any';
            }
        }
        else {
            type = 'Any';
        }
    }
    return Object.assign({}, rest, { type });
}
/**
 * ArrayBuffer.isView will only evaluate if an object instance is an
 * ArrayBufferView, but TypeScript metadata gives us a reference to the class.
 *
 * This function checks if the provided constructor is or extends the built-in
 * `ArrayBuffer` constructor, the `DataView` constructor, or any `TypedArray`
 * constructor.
 *
 * This function will need to be modified if new binary types are added to
 * JavaScript (e.g., the `Int64Array` or `Uint64Array` discussed in
 * {@link https://github.com/tc39/proposal-bigint the BigInt TC39 proposal}.
 *
 * @see https://developer.mozilla.org/en-US/docs/Web/API/ArrayBufferView
 */
function isBinaryType(arg) {
    return arg === Uint8Array || arg.prototype instanceof Uint8Array ||
        arg === Uint8ClampedArray || arg.prototype instanceof Uint8ClampedArray ||
        arg === Uint16Array || arg.prototype instanceof Uint16Array ||
        arg === Uint32Array || arg.prototype instanceof Uint32Array ||
        arg === Int8Array || arg.prototype instanceof Int8Array ||
        arg === Int16Array || arg.prototype instanceof Int16Array ||
        arg === Int32Array || arg.prototype instanceof Int32Array ||
        arg === Float32Array || arg.prototype instanceof Float32Array ||
        arg === Float64Array || arg.prototype instanceof Float64Array ||
        arg === ArrayBuffer || arg.prototype instanceof ArrayBuffer ||
        arg === DataView || arg.prototype instanceof DataView;
}

function hashKey(parameters = {}) {
    return attribute(Object.assign({}, parameters, { keyType: 'HASH' }));
}

function autoGeneratedHashKey(parameters = {}) {
    return hashKey(Object.assign({}, parameters, { type: 'String', defaultProvider: v4 }));
}

function rangeKey(parameters = {}) {
    return attribute(Object.assign({}, parameters, { keyType: 'RANGE' }));
}

/**
 * Declare a TypeScript class to be represent items in a table in a way
 * understandable by the AWS DynamoDB DataMapper for JavaScript. Meant to be
 * used as a TypeScript class decorator in projects compiled with the
 * `experimentalDecorators` option enabled.
 *
 * @see https://www.typescriptlang.org/docs/handbook/decorators.html
 * @see https://www.typescriptlang.org/docs/handbook/compiler-options.html
 */
function table(tableName) {
    return constructor => {
        constructor.prototype[DynamoDbTable] = tableName;
    };
}

function versionAttribute(parameters = {}) {
    return attribute(Object.assign({}, parameters, { type: 'Number', versionAttribute: true }));
}

export { attribute, autoGeneratedHashKey, hashKey, rangeKey, table, versionAttribute };
